@@ -31,7 +31,6 @@ int regexec_dfa(const regex_t *preg, const char *string, size_t nmatch,
     regmatch_t pmatch[], int /* eflags */)
 {
     const dfa_t *dfa = preg->dfa;
-    int result = REG_NOMATCH;
     regoff_t *regs = preg->regs;
     size_t i = 0;
     const char *p = string, *q = p;
@@ -62,46 +61,15 @@ int regexec_dfa(const regex_t *preg, const char *string, size_t nmatch,
         p = q;
     }
 
-    if (s->rule != Rule::NONE) {
-        regmatch_t *m = pmatch;
-        result = 0;
-        const regoff_t mlen = p - string - 1;
-
-        apply_regops(regs, s->tcmd[dfa->nchars], mlen);
-
-        m->rm_so = 0;
-        m->rm_eo = mlen;
-        ++m;
-
-        const Rule &rule = dfa->rules[0];
-        for (size_t t = rule.ltag; t < rule.htag; ++t) {
-
-            const Tag &tag = dfa->tags[t];
-            if (fictive(tag)) continue;
-            if (tag.ncap >= nmatch * 2) break;
-
-            regoff_t off;
-            if (!fixed(tag)) {
-                off = regs[dfa->finvers[t]];
-            }
-            else {
-                off = tag.base == Tag::RIGHTMOST
-                    ? mlen : regs[dfa->finvers[tag.base]];
-                DASSERT (off != -1);
-                off -= static_cast<regoff_t>(tag.dist);
-            }
-
-            if (tag.ncap % 2 == 0) {
-                m->rm_so = off;
-            }
-            else {
-                m->rm_eo = off;
-                ++m;
-            }
-        }
+    if (s->rule == Rule::NONE) {
+        return REG_NOMATCH;
     }
 
-    return result;
+    const regoff_t mlen = p - string - 1;
+    const getoff_dfa_t fn = { dfa, regs, mlen };
+    apply_regops(regs, s->tcmd[dfa->nchars], mlen);
+    tags_to_submatch(dfa->tags, nmatch, pmatch, mlen, fn);
+    return 0;
 }
 
 } // namespace libre2c
