@@ -138,7 +138,8 @@ int regexec_nfa_posix_backward(const regex_t *preg, const char *string
     ctx.step = len;
     ctx.cursor = ctx.marker = string + len;
 
-    const size_t sz = ctx.nfa.tags.size() * ctx.nfa.size * 2;
+    const std::vector<Tag> &tags = ctx.nfa.tags;
+    const size_t sz = tags.size() * ctx.nfa.size * 2;
     offsets4 = new regoff_t[sz];
     offsets5 = new regoff_t[sz];
     offsets6 = new regoff_t[sz];
@@ -164,20 +165,9 @@ int regexec_nfa_posix_backward(const regex_t *preg, const char *string
         return REG_NOMATCH;
     }
 
-    regmatch_t *m = pmatch;
-    m->rm_so = 0;
-    m->rm_eo = static_cast<regoff_t>(len);
-    const size_t n = std::min(ctx.nsub, 2 * nmatch);
-    for (size_t t = 0; t < n; ++t) {
-        const regoff_t off = ctx.offsets3[t];
-        if (t % 2 == 0) {
-            ++m;
-            m->rm_so = off;
-        }
-        else {
-            m->rm_eo = off;
-        }
-    }
+    const getoff_nfa_t fn = { ctx.offsets3 };
+    tags_to_submatch(tags, nmatch, pmatch, static_cast<regoff_t>(len), fn);
+
     return 0;
 }
 
@@ -565,11 +555,9 @@ void update_final_offsets(psimctx_t &ctx, const conf_t &c)
     if (D) fprintf(stderr, "finally: ");
     prtoff4(ctx, index(s, ctx.nfa));
 
-    for (size_t t = 0; t < ntags; ++t) {
-        const Tag &tag = tags[t];
-        if (!fictive(tag)) {
-            of[tag.ncap] = ox[2 * t + 1];
-        }
+    for (size_t t = 0; t < ntags; t += 2) {
+        of[t]     = ox[2 * t       + 1];
+        of[t + 1] = ox[2 * (t + 1) + 1];
     }
 }
 
