@@ -6,7 +6,7 @@
 #include <vector>
 
 #include "src/adfa/adfa.h"
-#include "src/codegen/go.h"
+#include "src/codegen/code.h"
 #include "src/debug/debug.h"
 #include "src/dfa/dfa.h"
 #include "src/options/opt.h"
@@ -17,8 +17,6 @@ namespace re2c {
 
 class Msg;
 
-const size_t Initial::NOSAVE = std::numeric_limits<size_t>::max();
-
 static bool is_eof(const opt_t *opts, uint32_t ub)
 {
     return opts->eof != NOEOF && static_cast<uint32_t>(opts->eof) == ub;
@@ -33,7 +31,7 @@ DFA::DFA
     , const std::string &nm
     , const std::string &cn
     , const std::string &su
-    , const Code *eof
+    , const SemAct *eof
     , const opt_t *opts
     , Msg &msg
     )
@@ -52,6 +50,10 @@ DFA::DFA
     , rules(dfa.rules)
     , tags(dfa.tags)
     , mtagvers(dfa.mtagvers)
+    , stagnames()
+    , stagvars()
+    , mtagnames()
+    , mtagvars()
     , finvers(dfa.finvers)
     , tcpool(dfa.tcpool)
     , max_fill (0)
@@ -62,10 +64,12 @@ DFA::DFA
     , maxtagver (dfa.maxtagver)
     , def_rule (def)
     , key_size (key)
-    , bitmaps (std::min(ubChar, 256u))
+    , bitmap(NULL)
     , setup(su)
     , eof_action(eof)
     , msg(msg)
+    , start_label(NULL)
+    , initial_label(NULL)
 {
     const size_t nstates = dfa.states.size();
     const size_t nchars = dfa.nchars;
@@ -113,7 +117,7 @@ DFA::DFA
             s->go.span[j].ub = charset[c];
             s->go.span[j].tags = tc;
         }
-        s->go.nSpans = j;
+        s->go.nspans = j;
     }
     *p = NULL;
 
@@ -172,7 +176,7 @@ void DFA::reorder()
         State *s = todo.front();
         todo.pop();
         ord.push_back(s);
-        for(uint32_t i = 0; i < s->go.nSpans; ++i)
+        for(uint32_t i = 0; i < s->go.nspans; ++i)
         {
             State *q = s->go.span[i].to;
             if(q && done.insert(q).second)
