@@ -9,9 +9,7 @@
 #include <vector>
 
 #include "src/adfa/action.h"
-#include "src/codegen/bitmap.h"
-#include "src/codegen/go.h"
-#include "src/codegen/label.h"
+#include "src/codegen/code.h"
 #include "src/dfa/tcmd.h"
 #include "src/msg/location.h"
 #include "src/regexp/rule.h"
@@ -25,12 +23,12 @@ class Msg;
 struct opt_t;
 class Output;
 struct dfa_t;
+struct Label;
 
-struct State
-{
-    label_t label;
+struct State {
     State * next;
     State * prev;
+    Label *label;
     size_t fill;
     bool fallback;
 
@@ -39,13 +37,13 @@ struct State
     tcid_t rule_tags;
     tcid_t fall_tags;
     bool isBase;
-    Go go;
+    CodeGo go;
     Action action;
 
-    State ()
-        : label (label_t::first ())
-        , next (0)
+    State()
+        : next (0)
         , prev (0)
+        , label(NULL)
         , fill (0)
         , fallback (false)
         , rule (Rule::NONE)
@@ -53,14 +51,15 @@ struct State
         , rule_tags (TCID0)
         , fall_tags (TCID0)
         , isBase (false)
-        , go ()
+        , go()
         , action ()
-    {}
-    ~State ()
+    {
+        init_go(&go);
+    }
+    ~State()
     {
         operator delete (go.span);
     }
-
     FORBID_COPY (State);
 };
 
@@ -81,6 +80,10 @@ struct DFA
     std::valarray<Rule> &rules;
     std::vector<Tag> &tags;
     std::set<tagver_t> &mtagvers;
+    std::set<std::string> stagnames;
+    std::set<std::string> stagvars;
+    std::set<std::string> mtagnames;
+    std::set<std::string> mtagvars;
     const tagver_t *finvers;
     tcpool_t &tcpool;
     size_t max_fill;
@@ -91,11 +94,14 @@ struct DFA
     tagver_t maxtagver;
     const size_t def_rule;
     const size_t key_size;
-    bitmaps_t bitmaps;
+    CodeBitmap *bitmap;
     std::string setup;
-    const Code *eof_action;
+    const SemAct *eof_action;
     Msg &msg;
 
+    Label *start_label;
+    Label *initial_label;
+
     DFA ( const dfa_t &dfa
         , const std::vector<size_t> &fill
         , size_t def
@@ -104,15 +110,16 @@ struct DFA
         , const std::string &nm
         , const std::string &cn
         , const std::string &su
-        , const Code *eof
+        , const SemAct *eof
         , const opt_t *opts
         , Msg &msg
         );
     ~DFA ();
     void reorder();
     void prepare(const opt_t *opts);
-    void calc_stats(const opt_t *opts);
-    void emit (Output &, uint32_t &, bool, bool &);
+    void calc_stats(OutputBlock &out);
+    void emit_body(Output &output, CodeList *program) const;
+    void emit_dot(Output &output, CodeList *program) const;
 
 private:
     void addState(State*, State *);
@@ -120,9 +127,6 @@ private:
     void findBaseState(const opt_t *opts);
     void hoist_tags(const opt_t *opts);
     void hoist_tags_and_skip(const opt_t *opts);
-    void count_used_labels(std::set<label_t> &used, label_t start, label_t initial, const opt_t *opts) const;
-    void emit_body (Output &, uint32_t &, const std::set<label_t> & used_labels, label_t initial) const;
-    void emit_dot(Output &o, bool last_cond) const;
 
     FORBID_COPY (DFA);
 };
