@@ -7,8 +7,7 @@
 
 #include "src/adfa/action.h"
 #include "src/adfa/adfa.h"
-#include "src/codegen/bitmap.h"
-#include "src/codegen/go.h"
+#include "src/codegen/code.h"
 #include "src/debug/debug.h"
 #include "src/dfa/tcmd.h"
 #include "src/msg/msg.h"
@@ -32,7 +31,7 @@ void DFA::split(State *s)
     move->rule_tags = s->rule_tags;
     move->fall_tags = s->fall_tags;
     s->rule = Rule::NONE;
-    s->go.nSpans = 1;
+    s->go.nspans = 1;
     s->go.span = allocate<Span> (1);
     s->go.span[0].ub = ubChar;
     s->go.span[0].to = move;
@@ -43,8 +42,8 @@ static uint32_t merge(Span *x, State *fg, State *bg, const opt_t *opts)
 {
     Span *f = fg->go.span;
     Span *b = bg->go.span;
-    Span *const fe = f + fg->go.nSpans;
-    Span *const be = b + bg->go.nSpans;
+    Span *const fe = f + fg->go.nspans;
+    Span *const be = b + bg->go.nspans;
     Span *const x0 = x;
     const uint32_t eofub = opts->eof + 1;
 
@@ -90,7 +89,7 @@ void DFA::findBaseState(const opt_t *opts)
 
     for (State *s = head; s; s = s->next) {
         if (s->fill == 0) {
-            for (uint32_t i = 0; i < s->go.nSpans; ++i) {
+            for (uint32_t i = 0; i < s->go.nspans; ++i) {
                 State *to = s->go.span[i].to;
 
                 if (to->isBase
@@ -99,13 +98,13 @@ void DFA::findBaseState(const opt_t *opts)
                     DASSERT(s->stadfa_tags == TCID0);
 
                     to = to->go.span[0].to;
-                    uint32_t nSpans = merge(span, s, to, opts);
+                    uint32_t nspans = merge(span, s, to, opts);
 
-                    if (nSpans < s->go.nSpans) {
+                    if (nspans < s->go.nspans) {
                         operator delete (s->go.span);
-                        s->go.nSpans = nSpans;
-                        s->go.span = allocate<Span> (nSpans);
-                        memcpy(s->go.span, span, nSpans*sizeof(Span));
+                        s->go.nspans = nspans;
+                        s->go.span = allocate<Span> (nspans);
+                        memcpy(s->go.span, span, nspans*sizeof(Span));
                         break;
                     }
                 }
@@ -163,7 +162,7 @@ void DFA::prepare(const opt_t *opts)
                 finstates[s->rule] = n;
                 addState(n, s);
             }
-            for (uint32_t i = 0; i < s->go.nSpans; ++i) {
+            for (uint32_t i = 0; i < s->go.nspans; ++i) {
                 if (!s->go.span[i].to) {
                     s->go.span[i].to = finstates[s->rule];
                     s->go.span[i].tags = s->rule_tags;
@@ -173,17 +172,12 @@ void DFA::prepare(const opt_t *opts)
     }
 
     // create default state (if needed)
-    State * default_state = NULL;
-    for (State * s = head; s; s = s->next)
-    {
-        for (uint32_t i = 0; i < s->go.nSpans; ++i)
-        {
-            if (!s->go.span[i].to)
-            {
-                if (!default_state)
-                {
-                    default_state = new State;
-                    defstate = default_state;
+    State *default_state = NULL;
+    for (State *s = head; s; s = s->next) {
+        for (uint32_t i = 0; i < s->go.nspans; ++i) {
+            if (!s->go.span[i].to) {
+                if (!default_state) {
+                    default_state = defstate = new State;
                     addState(default_state, s);
                 }
                 s->go.span[i].to = defstate;
@@ -191,11 +185,28 @@ void DFA::prepare(const opt_t *opts)
         }
     }
 
+    // With EOF rule there is a default quasi-transition on YYFILL failure, so
+    // default state is needed if there is at least one final state with at
+    // least one outgoing transition to a non-final state.
+    if (!default_state && opts->eof != NOEOF) {
+        bool have_fallback_states = false;
+
+        for (State *s = head; s; s = s->next) {
+            have_fallback_states |= s->fallback;
+
+            if (!s->next && have_fallback_states) {
+                default_state = defstate = new State;
+                addState(default_state, s);
+                break;
+            }
+        }
+    }
+
     // bind save actions to fallback states and create accept state (if needed)
     if (default_state) {
         for (State *s = head; s; s = s->next) {
             if (s->fallback) {
-                const std::pair<const State*, tcid_t> acc(finstates[s->rule], s->fall_tags);
+                const std::pair<State*, tcid_t> acc(finstates[s->rule], s->fall_tags);
                 s->action.set_save(accepts.find_or_add(acc));
             }
         }
@@ -214,15 +225,10 @@ void DFA::prepare(const opt_t *opts)
         s->isBase = false;
 
         if (s->fill != 0) {
-            for (uint32_t i = 0; i < s->go.nSpans; ++i) {
+            for (uint32_t i = 0; i < s->go.nspans; ++i) {
                 if (s->go.span[i].to == s) {
                     s->isBase = true;
                     split(s);
-
-                    if (opts->bFlag) {
-                        bitmaps.insert(&s->next->go, s);
-                    }
-
                     s = s->next;
                     break;
                 }
@@ -236,20 +242,16 @@ void DFA::prepare(const opt_t *opts)
     if (opts->eager_skip) {
         hoist_tags_and_skip(opts);
     }
-
-    for (State *s = head; s; s = s->next) {
-        s->go.init(s, opts, bitmaps);
-    }
 }
 
-void DFA::calc_stats(const opt_t *opts)
+void DFA::calc_stats(OutputBlock &out)
 {
+    const opt_t *opts = out.opts;
+
     // calculate 'YYMAXFILL'
     max_fill = 0;
-    for (State * s = head; s; s = s->next)
-    {
-        if (max_fill < s->fill)
-        {
+    for (State * s = head; s; s = s->next) {
+        if (max_fill < s->fill) {
             max_fill = s->fill;
         }
     }
@@ -284,12 +286,39 @@ void DFA::calc_stats(const opt_t *opts)
             "multiple context markers, use '-t, --tags' "
             "option and '/*!stags:re2c ... */' directive");
     }
+
+    if (!oldstyle_ctxmarker) {
+        for (size_t i = 0; i < tags.size(); ++i) {
+            const Tag &tag = tags[i];
+            if (history(tag)) {
+                mtagvars.insert(*tag.name);
+            }
+            else if (tag.name) {
+                stagvars.insert(*tag.name);
+            }
+        }
+        for (tagver_t v = 1; v <= maxtagver; ++v) {
+            const std::string s = vartag_name(v, opts->tags_prefix);
+            if (mtagvers.find(v) != mtagvers.end()) {
+                mtagnames.insert(s);
+            }
+            else {
+                stagnames.insert(s);
+            }
+        }
+        out.stags.insert(stagnames.begin(), stagnames.end());
+        out.mtags.insert(mtagnames.begin(), mtagnames.end());
+    }
+
+    if (!cond.empty()) {
+        out.types.push_back(cond);
+    }
 }
 
 static bool can_hoist_tags(const State *s, const opt_t *opts)
 {
     Span *span = s->go.span;
-    const size_t nspan = s->go.nSpans;
+    const size_t nspan = s->go.nspans;
     DASSERT(nspan != 0);
 
     if (nspan == 1 && s->rule != Rule::NONE) return false;
@@ -316,7 +345,7 @@ void DFA::hoist_tags(const opt_t *opts)
 {
     for (State * s = head; s; s = s->next) {
         Span *span = s->go.span;
-        const size_t nspan = s->go.nSpans;
+        const size_t nspan = s->go.nspans;
         if (nspan == 0) continue;
 
         if (can_hoist_tags(s, opts)) {
@@ -334,7 +363,7 @@ void DFA::hoist_tags_and_skip(const opt_t *opts)
 
     for (State * s = head; s; s = s->next) {
         Span *span = s->go.span;
-        const size_t nspan = s->go.nSpans;
+        const size_t nspan = s->go.nspans;
         if (nspan == 0) continue;
 
         // do all spans agree on tags?
