@@ -6,7 +6,7 @@
 #include <vector>
 
 #include "src/adfa/adfa.h"
-#include "src/codegen/output.h"
+#include "src/codegen/code.h"
 #include "src/compile.h"
 #include "src/debug/debug.h"
 #include "src/dfa/dfa.h"
@@ -22,23 +22,32 @@
 #include "src/skeleton/skeleton.h"
 #include "src/util/free_list.h"
 #include "src/util/range.h"
-#include "src/util/smart_ptr.h"
 
 
 namespace re2c {
 
 class Msg;
 
-static std::string make_name(const std::string &cond, uint32_t line)
+static std::string make_name(Output &output, const std::string &cond, const loc_t &loc)
 {
-    std::ostringstream os;
-    os << "line" << line;
-    std::string name = os.str();
-    if (!cond.empty ())
-    {
+    std::string name;
+
+    // if the block is included from another file, prepend filename for disambiguation
+    if (loc.file > 0) {
+        name += output.msg.filenames[loc.file];
+        for (size_t i = 0; i < name.length(); ++i) {
+            if (!std::isalnum(name[i])) name[i] = '_';
+        }
         name += "_";
-        name += cond;
     }
+
+    std::ostringstream os;
+    os << loc.line;
+    name += "line" + os.str();
+
+    // if the block has multiple conditions, append condition name for disambiguation
+    if (!cond.empty()) name += "_" + cond;
+
     return name;
 }
 
@@ -49,10 +58,10 @@ static smart_ptr<DFA> ast_to_dfa(const spec_t &spec, Output &output)
     Msg &msg = output.msg;
     const std::vector<ASTRule> &rules = spec.rules;
     const size_t defrule = spec.defs.empty() ? Rule::NONE : rules.size() - 1;
-    const Code *eof = spec.eofs.empty() ? NULL : spec.eofs.front();
+    const SemAct *eof = spec.eofs.empty() ? NULL : spec.eofs.front();
     const std::string
         &cond = spec.name,
-        name = make_name(cond, loc.line),
+        name = make_name(output, cond, loc),
         &setup = spec.setup.empty() ? "" : spec.setup[0]->text;
 
     RangeMgr rangemgr;
@@ -108,7 +117,7 @@ static smart_ptr<DFA> ast_to_dfa(const spec_t &spec, Output &output)
     DDUMP_ADFA(opts, *adfa);
 
     // finally gather overall DFA statistics
-    adfa->calc_stats(opts);
+    adfa->calc_stats(output.block());
 
     // accumulate global statistics from this particular DFA
     output.max_fill = std::max(output.max_fill, adfa->max_fill);
@@ -125,50 +134,49 @@ void compile(Scanner &input, Output &output, Opt &opts)
     specs_t rspecs;
     symtab_t symtab;
     const conopt_t *globopts = &opts.glob;
-    const opt_t *ropts = NULL;
-    typedef std::vector<smart_ptr<DFA> > dfas_t;
-
+    code_alc_t &alc = output.allocator;
     const loc_t &loc0 = input.tok_loc();
 
     output.header_mode(1);
-    output.new_block(opts, loc0);
+    output.new_block(opts, loc0, false /* reuse */);
     output.wversion_time();
 
     output.header_mode(0);
-    output.new_block(opts, loc0);
+    output.new_block(opts, loc0, false /* reuse */);
     output.wversion_time();
-    output.wdelay_line_info_input(loc0);
+    output.wdelay_stmt(0, code_line_info_input(alc, loc0));
 
     if (globopts->target == TARGET_SKELETON) {
-        emit_prolog(output);
+        output.wdelay_stmt(0, emit_skeleton_prolog(output));
     }
 
+    const opt_t *rules_opts = NULL;
+    const opt_t *accum_opts = output.block().opts;
+
     for (;;) {
         // parse everything up to the next re2c block
         Scanner::ParseMode mode = input.echo(output);
         if (mode == Scanner::Stop) break;
-        validate_mode(mode, globopts->rFlag, ropts, input);
+        validate_mode(mode, globopts->rFlag, rules_opts, input);
 
         // parse the next re2c block
         specs_t specs;
         if (mode == Scanner::Reuse) {
             specs = rspecs;
-            opts.restore(ropts);
-            opts.reset_mapCodeName();
-            output.fill_index = 0;
+            opts.restore(rules_opts);
             output.state_goto = false;
-            output.cond_goto = false;
         }
+        output.cond_goto = false;
         parse(input, specs, symtab, opts);
 
         // start new output block with accumulated options
         const loc_t &loc = input.cur_loc();
-        output.new_block(opts, loc);
+        output.new_block(opts, loc, mode == Scanner::Reuse);
 
         if (mode == Scanner::Rules) {
             // save AST and options for future use
             rspecs = specs;
-            ropts = output.block().opts;
+            rules_opts = output.block().opts;
         } else {
             validate_ast(specs, output.block().opts, output.msg);
             normalize_ast(specs);
@@ -180,22 +188,29 @@ void compile(Scanner &input, Output &output, Opt &opts)
             }
 
             // compile DFA to code
-            bool prolog = false;
-            uint32_t ind = output.block().opts->topIndent;
-            for (dfas_t::const_iterator i = dfas.begin(); i != dfas.end(); ++i) {
-                (*i)->emit(output, ind, (i + 1) == dfas.end(), prolog);
-            }
+            gen_code(output, dfas);
         }
+        output.wdelay_stmt(0, code_line_info_input(alc, loc));
 
-        output.wdelay_line_info_input(loc);
+        // accumulate whole-program information from this block
+        output.gather_info_from_block();
+
+        // accumulated whole-program options exclude rules/reuse blocks
+        if (mode == Scanner::Parse) {
+            accum_opts = output.block().opts;
+        } else {
+            opts.restore(accum_opts);
+        }
     }
 
+    output.total_opts = accum_opts ? accum_opts : rules_opts;
+
     if (globopts->target == TARGET_SKELETON) {
-        emit_epilog (output, output.skeletons);
+        output.wdelay_stmt(0, emit_skeleton_epilog(output));
     }
 
     AST::flist.clear();
-    Code::flist.clear();
+    SemAct::flist.clear();
     RangeSuffix::freeList.clear();
 }
 
