@@ -227,8 +227,8 @@ void warn_nondeterministic_tags(const ctx_t &ctx)
         for (size_t t = rule.ltag; t < rule.htag; ++t) {
             const size_t m = maxv[t];
             if (m > 1) {
-                warn.nondeterministic_tags(rule.code->loc
-                    , ctx.dc_condname, tags[t].name, m);
+                warn.nondeterministic_tags(rule.semact->loc, ctx.dc_condname,
+                    tags[t].name, m);
             }
         }
     }
@@ -273,6 +273,7 @@ determ_context_t<history_t>::determ_context_t(const opt_t *opts, Msg &msg
     , sortcores()
     , fincount()
     , worklist()
+    , dump_dfa_tree(*this)
     , dc_dump(opts)
     , dc_clstats()
 {
