@@ -30,9 +30,9 @@ void normalize_ast(specs_t &specs)
     // merge default rule with the lowest priority
     for (i = specs.begin(), e = specs.end(); i != e; ++i) {
         if (!i->defs.empty()) {
-            const Code *c = i->defs[0];
-            const AST *r = ast_default(c->loc);
-            i->rules.push_back(ASTRule(r, c));
+            const SemAct *a = i->defs[0];
+            const AST *r = ast_default(a->loc);
+            i->rules.push_back(ASTRule(r, a));
         }
     }
 
