@@ -62,7 +62,7 @@ static Range *dot_to_range(RESpec &, const AST *);
 static Range *cls_to_range(RESpec &, const AST *);
 static bool misuse_of_named_def(RESpec &, const AST *);
 static void assert_tags_used_once(RESpec &, const Rule &, const std::vector<Tag> &);
-static void init_rule(RESpec &, Rule &, const Code *, const std::vector<Tag> &, size_t, size_t);
+static void init_rule(RESpec &, Rule &, const SemAct *, const std::vector<Tag> &, size_t, size_t);
 static bool is_icase(const opt_t *, bool);
 
 
@@ -80,7 +80,7 @@ RESpec::RESpec(const std::vector<ASTRule> &ast, const opt_t *o, Msg &msg
     for (size_t i = 0; i < ast.size(); ++i) {
         size_t ltag = tags.size(), ncap = 0;
         res.push_back(ast_to_re(*this, ast[i].ast, ncap, 0));
-        init_rule(*this, rules[i], ast[i].code, tags, ltag, ncap);
+        init_rule(*this, rules[i], ast[i].semact, tags, ltag, ncap);
     }
 }
 
@@ -103,6 +103,52 @@ bool has_tags(const AST *ast)
     return false; /* unreachable */
 }
 
+static inline void add_fictive_tags(RESpec &spec, int32_t height,
+    RE **ptag1, RE **ptag2)
+{
+    std::vector<Tag> &tags = spec.tags;
+
+    // opening fictive tag
+    *ptag1 = re_tag(spec, tags.size(), false);
+    tags.push_back(Tag(Tag::FICTIVE, Tag::FICTIVE, false, height + 1));
+
+    // closing fictive tag
+    *ptag2 = re_tag(spec, tags.size(), false);
+    tags.push_back(Tag(Tag::FICTIVE, Tag::FICTIVE, false, height));
+}
+
+static inline void add_capture_tags(RESpec &spec, const AST **past,
+    size_t &ncap, int32_t height, RE **ptag1, RE **ptag2, bool orbit)
+{
+    std::vector<Tag> &tags = spec.tags;
+    const size_t lcap = ncap;
+    const AST *ast = *past;
+
+    DASSERT(ast->type == AST::CAP);
+
+    // save the range of repeated captures and collapse them: (...(R)...) -> (R)
+    for (ast = ast->cap; ast && ast->type == AST::CAP; ast = ast->cap) {
+        ++ncap;
+    }
+
+    // dereference to avoid future check for non-parenthesized rerefences
+    if (ast->type == AST::REF) {
+        ast = ast->ref.ast;
+    }
+
+    *past = ast;
+
+    // opening capture tag
+    *ptag1 = re_tag(spec, tags.size(), false);
+    tags.push_back(Tag(2 * lcap, 2 * ncap, orbit, height + 1));
+
+    // closing capture tag
+    *ptag2 = re_tag(spec, tags.size(), false);
+    tags.push_back(Tag(2 * lcap + 1, 2 * ncap + 1, orbit, height));
+
+    ++ncap;
+}
+
 RE *ast_to_re(RESpec &spec, const AST *ast, size_t &ncap, int32_t height)
 {
     std::vector<Tag> &tags = spec.tags;
@@ -113,77 +159,71 @@ RE *ast_to_re(RESpec &spec, const AST *ast, size_t &ncap, int32_t height)
     switch (ast->type) {
         case AST::NIL:
             return re_nil(spec);
+
         case AST::STR:
             return re_string(spec, ast);
+
         case AST::CLS: {
             Range *r = cls_to_range(spec, ast);
             return re_class(spec, ast->loc, r);
         }
+
         case AST::DOT: {
             Range *r = dot_to_range(spec, ast);
             return re_class(spec, ast->loc, r);
         }
+
         case AST::DEFAULT: {
             // see note [default regexp]
             Range *r = spec.rangemgr.ran(0, opts->encoding.nCodeUnits());
             return re_sym(spec, r);
         }
+
         case AST::DIFF: {
             Range *r = diff_to_range(spec, ast);
             return re_class(spec, ast->loc, r);
         }
+
         case AST::ALT: {
             // see note [POSIX subexpression hierarchy]
             const bool need_tags = opts->posix_semantics && has_tags(ast);
-
             RE *t1 = NULL, *t2 = NULL, *t3 = NULL, *t4 = NULL, *x, *y;
 
             if (need_tags && ast->alt.ast1->type != AST::CAP) {
-                t1 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height + 1));
-                t2 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height));
+                add_fictive_tags(spec, height, &t1, &t2);
             }
             x = ast_to_re(spec, ast->alt.ast1, ncap, height);
             x = re_cat(spec, t1, re_cat(spec, x, t2));
 
             if (need_tags && ast->alt.ast2->type != AST::CAP) {
-                t3 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height + 1));
-                t4 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height));
+                add_fictive_tags(spec, height, &t3, &t4);
             }
             y = ast_to_re(spec, ast->alt.ast2, ncap, height);
             y = re_cat(spec, t3, re_cat(spec, y, t4));
 
             return re_alt(spec, x, y);
         }
+
         case AST::CAT: {
             // see note [POSIX subexpression hierarchy]
             const bool need_tags = opts->posix_semantics && has_tags(ast);
-
             RE *t1 = NULL, *t2 = NULL, *t3 = NULL, *t4 = NULL, *x, *y;
 
             if (need_tags && ast->cat.ast1->type != AST::CAP) {
-                t1 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height + 1));
-                t2 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height));
+                add_fictive_tags(spec, height, &t1, &t2);
             }
             x = ast_to_re(spec, ast->cat.ast1, ncap, height);
             x = re_cat(spec, t1, re_cat(spec, x, t2));
 
             if (need_tags && ast->cat.ast2->type != AST::CAP) {
-                t3 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height + 1));
-                t4 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(Tag::FICTIVE, false, height));
+                add_fictive_tags(spec, height, &t3, &t4);
             }
             y = ast_to_re(spec, ast->cat.ast2, ncap, height);
             y = re_cat(spec, t3, re_cat(spec, y, t4));
 
             return re_cat(spec, x, y);
         }
+
         case AST::TAG: {
             if (ast->tag.name && !opts->tags) {
                 spec.msg.fatal(ast->loc
@@ -197,55 +237,43 @@ RE *ast_to_re(RESpec &spec, const AST *ast, size_t &ncap, int32_t height)
             tags.push_back(Tag(ast->tag.name, ast->tag.history, height));
             return t;
         }
-        case AST::CAP: {
+
+        case AST::CAP:
             if (!opts->posix_syntax) {
                 return ast_to_re(spec, ast->cap, ncap, height);
             }
-            const AST *x = ast->cap;
-            if (x->type == AST::REF) x = x->ref.ast;
+            else {
+                RE *y = NULL, *t1 = NULL, *t2 = NULL;
+                add_capture_tags(spec, &ast, ncap, height, &t1, &t2, false);
+                y = ast_to_re(spec, ast, ncap, height);
+                return re_cat(spec, t1, re_cat(spec, y, t2));
+            }
 
-            RE *t1 = re_tag(spec, tags.size(), false);
-            tags.push_back(Tag(2 * ncap, false, height + 1));
-
-            RE *t2 = re_tag(spec, tags.size(), false);
-            tags.push_back(Tag(2 * ncap + 1, false, height));
-
-            ++ncap;
-            RE *y = ast_to_re(spec, x, ncap, height);
-            return re_cat(spec, t1, re_cat(spec, y, t2));
-        }
         case AST::REF:
-            if (misuse_of_named_def(spec, ast)) return NULL;
+            if (misuse_of_named_def(spec, ast)) {
+                return NULL;
+            }
             return ast_to_re(spec, ast->ref.ast, ncap, height);
+
         case AST::ITER: {
-            const uint32_t
-                n = ast->iter.min,
-                n1 = std::max(n, 1u),
-                m = std::max(n, ast->iter.max);
-            const AST *x = ast->iter.ast;
-
-            RE *t1 = NULL, *t2 = NULL;
-            if (opts->posix_semantics && x->type == AST::CAP) {
-                x = x->cap;
-                if (x->type == AST::REF) x = x->ref.ast;
-
-                t1 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(2 * ncap, m > 1, height + 1));
+            const uint32_t n = ast->iter.min;
+            const uint32_t n1 = std::max(n, 1u);
+            const uint32_t m = std::max(n, ast->iter.max);
+            RE *y = NULL, *t1 = NULL, *t2 = NULL;
 
-                t2 = re_tag(spec, tags.size(), false);
-                tags.push_back(Tag(2 * ncap + 1, m > 1, height));
+            ast = ast->iter.ast;
 
-                ++ncap;
+            if (opts->posix_semantics && ast->type == AST::CAP) {
+                add_capture_tags(spec, &ast, ncap, height, &t1, &t2, m > 1);
             }
 
-            RE *y = NULL;
             if (m == 0) {
                 y = re_cat(spec, t1, t2);
             } else if (m == 1) {
-                y = ast_to_re(spec, x, ncap, height);
+                y = ast_to_re(spec, ast, ncap, height);
                 y = re_cat(spec, t1, re_cat(spec, y, t2));
             } else  {
-                y = ast_to_re(spec, x, ncap, height);
+                y = ast_to_re(spec, ast, ncap, height);
                 y = re_cat(spec, t1, y);
                 y = re_cat(spec, y, t2);
                 y = re_iter(spec, y, n1, m);
@@ -253,9 +281,11 @@ RE *ast_to_re(RESpec &spec, const AST *ast, size_t &ncap, int32_t height)
             if (n == 0) {
                 y = re_alt(spec, y, re_nil(spec));
             }
+
             return y;
         }
     }
+
     return NULL; /* unreachable */
 }
 
@@ -426,17 +456,17 @@ void assert_tags_used_once(RESpec &spec, const Rule &rule
     for (size_t t = rule.ltag; t < rule.htag; ++t) {
         name = tags[t].name;
         if (name && !names.insert(*name).second) {
-            spec.msg.fatal(rule.code->loc
+            spec.msg.fatal(rule.semact->loc
                 , "tag '%s' is used multiple times in the same rule"
                 , name->c_str());
         }
     }
 }
 
-void init_rule(RESpec &spec, Rule &rule, const Code *code
+void init_rule(RESpec &spec, Rule &rule, const SemAct *semact
     , const std::vector<Tag> &tags, size_t ltag, size_t ncap)
 {
-    rule.code = code;
+    rule.semact = semact;
     rule.ltag = ltag;
     rule.htag = tags.size();
     for (rule.ttag = ltag; rule.ttag < rule.htag && !trailing(tags[rule.ttag]); ++rule.ttag);
