@@ -11,6 +11,7 @@
 #include <valarray>
 #include <vector>
 
+#include "src/codegen/code.h"
 #include "src/msg/location.h"
 #include "src/regexp/rule.h"
 #include "src/regexp/tag.h"
@@ -21,7 +22,6 @@
 
 namespace re2c {
 
-class Output;
 class path_t;
 struct DFA;
 struct dfa_state_t;
@@ -109,16 +109,11 @@ uint32_t maxpath(const Skeleton &skel);
 void warn_undefined_control_flow(const Skeleton &skel);
 void fprint_default_path(FILE *f, const Skeleton &skel, const path_t &p);
 void emit_data(const Skeleton &skel);
-void emit_prolog(Output & o);
-void emit_start(Output &o, size_t maxfill, size_t maxnmatch, const std::string &name,
-    size_t sizeof_key, size_t def, bool backup, bool accept, bool oldstyle_ctxmarker,
-    const std::set<std::string> &stagnames, const std::set<std::string> &stagvars,
-    const std::set<std::string> &mtagnames, const std::set<std::string> &mtagvars,
-    bitmaps_t &bitmaps);
-void emit_end(Output &o, const std::string &name, bool backup, bool oldstyle_ctxmarker,
-    const std::set<std::string> &mtagnames);
-void emit_epilog(Output &o, const std::set<std::string> &names);
-void emit_action(Output &o, uint32_t ind, const DFA &dfa, size_t rid);
+Code *emit_skeleton_prolog(Output &output);
+Code *emit_skeleton_epilog(Output &output);
+void emit_skeleton(Output &output, CodeList *code, DFA &dfa);
+void emit_skeleton_action(Output &output, CodeList *code, const DFA &dfa, size_t rid);
+void emit_skeleton_action_eof(Output &output, CodeList *code);
 
 } // namespace re2c
 
